%%
% NLDelayExample.m
%
% M-file to analyze a nonlinear delayed system using IQCs and dissipation
% inequalities.  The delayed system is a classical negative feedback with
% a nonlinear loop system L. A constant delay exists at the input of L.
% This file first analyzes the linearized closed-loop for both stability
% (delay margin) and performance (induced L2 gain). Then the stability
% and performance of the nonlinear system is analyzed.

%%
% Select the IQC multipliers to be used for analysis.
% The IQC multipliers are constructed using the helper function DELAYIQC
% based on the following types:
%     -type==1: Circle IQC with center = -1 and radius = 1.
%     -type==2: Circle IQC with center = 0 and radius = |phi|.
%     -type==3: Circle IQC with center = |phi| and radius = |phi|.
% The multipliers can be weighted by frequency since the delay is constant.
iqccase = 3;
PI = ss([]);
switch iqccase
    case 1
        IQCtype = [1; 2];
        IQCwt = [1; 1];
    case 2
        IQCtype = [1; 2];
        IQCwt = [1; tf([0.1 0.1],[1 0.1])];
    case 3
        IQCtype = [1; 3];
        IQCwt = [1; 1];
end

%% Linear Analysis: Delay Margin
% Compute the delay margin for the negative feedback system with loop
% transfer function L(s)*exp(-s*tau).

% Loop transfer function
A = [-49 0; 1 0];
B = [8; 0];
C = [-4.5 1.5];
D = 0;
L = ss(A,B,C,D);

% Perform loop-shift to represent reference to error transfer function
% as Fu(Glin,Stau) where Stau(v) := Dtau(v)-v. Glin is [-T S; -T S] where
% S/T are the sensitivity/complementary sensitivity transfer functions.
% The Matlab expression below gives a minimal realization for Glin.
Glin = ss(A-B*C,[B B],[-C; -C],[0 1; 0 1]);

% Determine exact delay margin and delay margin frequency.
am = allmargin(L);
[~,idx] = min( am.DelayMargin );
DelayMargin = am.DelayMargin(idx)
DMFrequency = am.DMFrequency(idx)

% Compute lower bound on delay margin using IQCs and bisection
taulb = 0;
tauub = 5;
tol = 1e-3;
infoLIN = [];
while tauub-taulb>tol
    % Delay value to analyze
    tautry = (taulb+tauub)/2;
    
    % Specify IQCs
    PI(:,:,1) = delayiqc(IQCtype(1),tautry,IQCwt(1));
    PI(:,:,2) = delayiqc(IQCtype(2),tautry,IQCwt(2));
    
    % Compute IQC upper bound on gain from reference to error
    [giqc,info] = lpviqcgain(Glin,PI);
    
    % System is input/output stable if gain is finite
    if isempty(giqc)
        tauub = tautry;
    else
        taulb = tautry;
        infoLIN = info;
    end
end
LinDelayMarginLB = taulb

%% Linear Analysis: Performance
% Compute the gain from reference to error for the negative feedback
% system with loop transfer function L(s)*exp(-s*tau).

% Grid of delay values
Ntau = 20;
taug = linspace(0.01,0.99,Ntau)*LinDelayMarginLB;

% Evaluate exact gain and IQC upper bound
Nw = 200;
w = logspace(-2,2,Nw)*DMFrequency;
Dtau = ss(1);
g = zeros(Ntau,1);
gUB = zeros(Ntau,1);
for i=1:Ntau
    % Set delay value
    tautry = taug(i);
    
    % Gain from ref to error based on frequency response
    Dtau.InputDelay = tautry;
    F = lft(Dtau-1,Glin);
    g(i) = norm( frd(F,w), inf);
    
    % Compute gain upper bound using IQCs
    PI(:,:,1) = delayiqc(IQCtype(1),tautry,IQCwt(1));
    PI(:,:,2) = delayiqc(IQCtype(2),tautry,IQCwt(2));
    [gUB(i),info] = lpviqcgain(Glin,PI);
end

%% Nonlinear Analysis: Delay Margin
% Compute the delay margin for the negative feedback system with loop
% L*exp(-s*tau) where L is a nonlinear system.

% Nonlinear loop system, L
%   d(xG)/dt = A*xG + B*wtil + nonlinear terms :=fL(x,wtil)
%   y = C*xG
pvar d w wtil;
ns = size(A,1);
xG = mpvar('xG',[ns 1]);

fL = A*xG + B*wtil + [2*xG(1)^2 + 3*xG(2)^2 - 0.2*xG(1)^3; -1*xG(2)^3];
y = C*xG;

% Perform loop-shift to represent reference to error as Fu(Gnl,Stau).
% Gnl has the form:
%    d(xG)/dt = f(xG,w,d)
%           v = h1(xG,w,d)
%           e = h2(xG,w,d)
f = subs(fL,wtil,w+d-y);
h1 = d-y;
h2 = d-y;
h = [h1;h2];

% Compute lower bound on delay margin using IQCs and bisection
taulb = 0;
tauub = 5;
tol = 1e-3;
infoNL = [];
cnt = 0;
tic
while tauub-taulb>tol
    % Delay value to analyze
    tautry = (taulb+tauub)/2;
    
    % Specify IQCs
    PI(:,:,1) = delayiqc(IQCtype(1),tautry,IQCwt(1));
    PI(:,:,2) = delayiqc(IQCtype(2),tautry,IQCwt(2));
    
    % Compute IQC upper bound on gain from reference to error
    [giqclmi,info] = nliqcgain(f,h,xG,w,d,PI);
    
    % System is input/output stable if gain is finite
    if isempty(giqclmi)
        tauub = tautry;
    else
        taulb = tautry;
        infoNL = info;
    end
    cnt = cnt+1;
end
tNL = toc;
disp(['Comp Time for NL Delay Margin = ' num2str(tNL)]);
disp(['# of Bisection Steps = ' int2str(cnt)]);
NLDelayMarginLB = taulb

%% Nonlinear Analysis: Performance
% Compute the gain from reference to error for the negative feedback
% system with nonlinear system L

% Grid of delay values
Ntau = 20;
taugNL = linspace(0.01,0.99,Ntau)*NLDelayMarginLB;

% Evaluate IQC upper bound
gNLUB = zeros(Ntau,1);
tic
for i=1:Ntau
    % Set delay value
    tautry = taugNL(i);
    
    % Compute gain upper bound using IQCs
    PI(:,:,1) = delayiqc(IQCtype(1),tautry,IQCwt(1));
    PI(:,:,2) = delayiqc(IQCtype(2),tautry,IQCwt(2));
    [gNLUB(i),info] = nliqcgain(f,h,xG,w,d,PI);
end
tNL2 = toc;
disp(['Comp Time for NL L2 Gain = ' num2str(tNL2)]);

%% 
% Plots of results
if iqccase==3
    NLPaperPlots
end