function [g,info] = nliqcgain(f,h,xG,w,d,PI)
% function [g,info] = nliqcgain(f,h,w,d,PI)
%
% Compute bound on gain of Fu(G,Delta) where G is a nonlinear system and 
% Delta is an operator described by IQCs PI. G is a described by
% a finite-dimensional ODE with polynomial vector fields. Thus the
% interconnection Fu(G,Delta) is given by:
%       dxG/dt = f(xG,w,d)
%            v = h1(xG,w,d)
%            e = h2(xG,w,d)
%            w = Delta(v)
%
% Inputs
%  f: Polynomial vector field for dynamics of G. (Ns-by-1 polynomial)
%  h: Polynomial output equation for G. ( (Nv+Ne)-by-1 polynomial )
%  xG: State of G specified as an Ns-by-1 vector of pvars
%  w: Outputs from Delta specified as an Nw-by-1 vector of pvars.
%  d: Input disturbances to G specified as an Nd-by-1 vector of pvars.
%  PI:  Parahermitian IQC multipliers as a TF/SS object. PI(:,:,i) is
%       the i^th multiplier for Delta. This function assumes Delta
%       is square m-by-m and hence PI should be m-by-m-Niqc where
%       Niqc is the number of IQCs.
%
% Outputs
%  g: Bound on the gain of Fu(G,Delta)
%  info:  Structure with information regarding solution of dissipation
%         inequality. Structure includes the fields:
%    -P:  Symmetric matrix used to construct the storage function.
%    -lamopt: Coefficients for optimal conic combination of IQC multipliers
%    -PIopt: Optimal multiplier obtained as the conic combination
%          PIopt = sum_i lam(i) PI(:,:,i)
%

% 12/23/2013 PJS Initial Coding

% XXX PJS: Add error checking
% Consistent dimensions
% Variable names consistent between input arguments.

%% Dimensions
nw = numel(w); % # of outputs of Delta 
nv = nw;       % # of inputs to Delta (Delta assumed to be square)
nd = numel(d); % # of disturbances
nwd = nw+nd;   % # of inputs of G = nw + nd
nve = numel(h);  % # of outputs of G = nv + ne
ne = nve-nv;    % # of errors

Niqc = size(PI,3);    % # of IQCs

%% Compute hard factorization for each IQC multiplier
Psi = cell(Niqc,1);
M = cell(Niqc,1);
for i=1:Niqc
    [Psi{i},M{i}] = jspec(PI(:,:,i));
end

%%
% Construct extended system Pext including dynamics of plant G and
% all auxiliary systems Psi. The I/O relationship for Pext is:
%     [z1; ...; zNiqc; e] = Pext*[w; d]
% where zj is the output of the auxiliary system Psi for the j^th IQC.

% Stack all auxiliary systems:
%      [z1; ...; zNiqc] = AllPsi*[v; w]
% zidx is a cell array of the output indices for zi.
AllPsi = vertcat(Psi{:});
nz = size(AllPsi,1);

zptr = 0;
zidx = cell(Niqc,1);
for i=1:Niqc
    zidx{i} = zptr + (1:2*nv)'; 
    zptr = zptr + 2*nv;
end

% Partition outputs of G correponding to v and e.
v = h(1:nv);
e = h(nv+1:end);

% Express dynamics of AllPsi using polynomial objects
% XXX -- This assumes that the variable 'psi' is not used as
% to name the state and/or inputs of G. Add error checking here.
[Apsi,Bpsi,Cpsi,Dpsi] = ssdata(AllPsi);
npsi = size(Apsi,1);
psi = mpvar('psi',[npsi,1]);
psidot = Apsi*psi + Bpsi*[v;w];
z = Cpsi*psi + Dpsi*[v;w];

% Form extended system as described by an ODE
%        dx/dt = F(x,w,d)
%        [z;e] = H(x,w,d)
% where x:=[xG;psi] is the combined state of G and all IQC filters Psi.
F = [f; psidot];
H = [z; e];
x = [xG; psi];

%%
% Specify SOS Constraints

% Initialize decision variables for V, lam, and gain.
% XXX -- Need to think about the monomials to include in V. AllPsi is
% LTI and, based on simple examples, it does not seem that higher order
% monomials in the state psi are requried.
pvar gsq;
lam = mpvar('lam',[Niqc,1]);
if 1
    % Include all quadratic and quartic monomials
    [V,P,q] = sosdecvar('cV',monomials(x,1:2));
else
    % Incllude all quadratic monomials but only quartic monomials in xG
    [V,P,q] = sosdecvar('cV',[x; monomials(xG,2)]);
end

% Form dissipation inequality DI>=0
DI = gsq*d'*d - e'*e - jacobian(V,x)*F;
for i=1:Niqc
    zi = z(zidx{i});
    Mi = M{i};
    DI = DI - lam(i)*(zi'*Mi*zi);
end

% Construct all SOS constraints
obj = gsq;
constraint1 =  V>=0;
constraint2 =  gsq>=0;
constraint3 = lam>=0;
constraint4 =  DI >=0;
constraints = [constraint1; constraint2; constraint3; constraint4];

%%
% Solve SOS optimization
opt = sosoptions;
% opt.solveropts.fid = 1;
% opt.simplify = 'off';
[sosinfo,dopt,sossol] = sosopt(constraints,[x;w;d],obj,opt);

%%
% Store solution
if ~isempty(dopt)
    g = sqrt(double(subs(gsq,dopt)));
    info.P = double(subs(P,dopt));
    info.V = subs(V,dopt);
    info.q = q;
    info.lamopt = double(subs(lam,dopt));
    PIopt = ss(zeros(nv+nw,nv+nw));
    for i = 1:Niqc
        PIopt = PIopt + info.lamopt(i)*PI(:,:,i);
    end
else
    g = [];
    info.P = [];
    info.V = [];
    info.lamopt = [];
    info.PIopt = [];    
end
info.sosinfo = sosinfo;
info.dopt = dopt;
info.sossol = sossol;

